@@ -30,10 +30,11 @@
 #include "gcin-im-client.h"
 #include <X11/keysym.h>
 
-// #define NEW_GTK_IM 1
+//#define NEW_GTK_IM 0
 #define DBG 0
 
 typedef struct _GtkGCINInfo GtkGCINInfo;
+//gboolean is_gdm;
 
 struct _GtkIMContextGCIN
 {
@@ -43,14 +44,18 @@ struct _GtkIMContextGCIN
   GtkWidget *client_widget;
   GCIN_client_handle *gcin_ch;
   int timeout_handle;
-  gboolean is_mozilla, dirty_fix_off;
+  gboolean is_mozilla, dirty_fix_off, preedit;
+  char *pe_str;
+  GCIN_PREEDIT_ATTR *pe_att;
+  int pe_attN;
+  int pe_cursor;
 };
 
 
 #if NEW_GTK_IM
 static void cancel_timeout(GtkIMContextGCIN *context)
 {
-#if DBG
+#if DBG && 0
   printf("cancel_timeout %d\n", context->timeout_handle);
 #endif
   if (!context->timeout_handle)
@@ -71,13 +76,13 @@ static void     gtk_im_context_gcin_reset              (GtkIMContext          *c
 static void     gtk_im_context_gcin_focus_in           (GtkIMContext          *context);
 static void     gtk_im_context_gcin_focus_out          (GtkIMContext          *context);
 static void     gtk_im_context_gcin_set_cursor_location (GtkIMContext          *context,
-						       GdkRectangle		*area);
+                                                       GdkRectangle             *area);
 static void     gtk_im_context_gcin_set_use_preedit    (GtkIMContext          *context,
-						       gboolean		      use_preedit);
+                                                       gboolean               use_preedit);
 static void     gtk_im_context_gcin_get_preedit_string (GtkIMContext          *context,
-						       gchar                **str,
-						       PangoAttrList        **attrs,
-						       gint                  *cursor_pos);
+                                                       gchar                **str,
+                                                       PangoAttrList        **attrs,
+                                                       gint                  *cursor_pos);
 
 // static GObjectClass *parent_class;
 
@@ -101,9 +106,9 @@ gtk_im_context_gcin_register_type (GTypeModule *type_module)
 
   gtk_type_im_context_gcin =
     g_type_module_register_type (type_module,
-				 GTK_TYPE_IM_CONTEXT,
-				 "GtkIMContextGCIN",
-				 &im_context_gcin_info, 0);
+                                 GTK_TYPE_IM_CONTEXT,
+                                 "GtkIMContextGCIN",
+                                 &im_context_gcin_info, 0);
 }
 
 static void
@@ -116,7 +121,7 @@ reinitialize_all_ics (GtkGCINInfo *info)
 
 #if 0
 static void gcin_display_closed (GdkDisplay *display,
-			 gboolean    is_error,
+                         gboolean    is_error,
                          GtkIMContextGCIN *context_xim)
 {
 #if DBG
@@ -133,12 +138,6 @@ static void gcin_display_closed (GdkDisplay *display,
 }
 #endif
 
-static void
-preedit_style_change (GtkGCINInfo *info)
-{
-//  dbg("preedit_style_change\n");
-}
-
 
 static void
 get_im (GtkIMContextGCIN *context_xim)
@@ -189,6 +188,12 @@ static void
 gtk_im_context_gcin_init (GtkIMContextGCIN *im_context_gcin)
 {
   im_context_gcin->timeout_handle = 0;
+  char *pe_str;
+  im_context_gcin->pe_attN = 0;
+  im_context_gcin->pe_att = NULL;
+  im_context_gcin->pe_str = NULL;
+  im_context_gcin->pe_cursor = 0;
+
 #if DBG
   printf("gtk_im_context_gcin_init %x\n", im_context_gcin);
 #endif
@@ -196,8 +201,8 @@ gtk_im_context_gcin_init (GtkIMContextGCIN *im_context_gcin)
   int pid = getpid();
 // probably only works for linux
   static char *moz[]={"mozilla", "firefox", "thunderbird", "nvu", "sunbird",
-	"seamonkey", "gnuzilla", "iceweasel", "icedove", "iceape", "swiftfox",
-	"iceowl", "kompozer", "swiftdove", "swiftweasel", "navigator", "xulrunner"};
+        "seamonkey", "gnuzilla", "iceweasel", "icedove", "iceape", "swiftfox",
+        "iceowl", "kompozer", "swiftdove", "swiftweasel", "navigator", "xulrunner"};
   char tstr0[64];
   char exec[256];
   sprintf(tstr0, "/proc/%d/exe", pid);
@@ -211,11 +216,32 @@ gtk_im_context_gcin_init (GtkIMContextGCIN *im_context_gcin)
       im_context_gcin->is_mozilla = TRUE;
       break;
     }
+
+//    if (strstr(exec, "/gdm"))
+//      is_gdm = TRUE;
   }
 #endif
 // dirty hack for mozilla...
 }
 
+
+void clear_preedit(GtkIMContextGCIN *context_gcin)
+{
+  if (context_gcin->pe_str) {
+    free(context_gcin->pe_str);
+    context_gcin->pe_str = NULL;
+  }
+
+  if (context_gcin->pe_att) {
+    free(context_gcin->pe_att);
+    context_gcin->pe_att = NULL;
+    context_gcin->pe_attN = 0;
+  }
+
+  context_gcin->pe_cursor = 0;
+}
+
+
 static void
 gtk_im_context_gcin_finalize (GObject *obj)
 {
@@ -223,6 +249,7 @@ gtk_im_context_gcin_finalize (GObject *obj)
   printf("gtk_im_context_gcin_finalize %x\n", obj);
 #endif
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (obj);
+  clear_preedit(context_xim);
 
 #if NEW_GTK_IM
   cancel_timeout(context_xim);
@@ -246,7 +273,7 @@ widget_for_window (GdkWindow *window)
       gpointer user_data;
       gdk_window_get_user_data (window, &user_data);
       if (user_data)
-	return user_data;
+        return user_data;
 
       window = gdk_window_get_parent (window);
     }
@@ -261,8 +288,8 @@ static void update_in_toplevel (GtkIMContextGCIN *context_xim)
 }
 static void
 on_client_widget_hierarchy_changed (GtkWidget       *widget,
-				    GtkWidget       *old_toplevel,
-				    GtkIMContextGCIN *context_xim)
+                                    GtkWidget       *old_toplevel,
+                                    GtkIMContextGCIN *context_xim)
 {
   update_in_toplevel (context_xim);
 }
@@ -280,37 +307,12 @@ set_ic_client_window (GtkIMContextGCIN *context_xim,
 
   context_xim->client_window = client_window;
 
-  if (context_xim->client_window)
-    {
-      get_im (context_xim);
-      if (context_xim->gcin_ch) {
-        gcin_im_client_set_window(context_xim->gcin_ch, GDK_DRAWABLE_XID(client_window));
-      }
+  if (context_xim->client_window) {
+    get_im (context_xim);
+    if (context_xim->gcin_ch) {
+      gcin_im_client_set_window(context_xim->gcin_ch, GDK_DRAWABLE_XID(client_window));
     }
-
-#if 0
-  // cause coredump
-  GtkWidget *new_client_widget = widget_for_window (context_xim->client_window);
-
-  if (new_client_widget != context_xim->client_widget)
-    {
-      if (context_xim->client_widget)
-	{
-	  g_signal_handlers_disconnect_by_func (context_xim->client_widget,
-						G_CALLBACK (on_client_widget_hierarchy_changed),
-						context_xim);
-	}
-      context_xim->client_widget = new_client_widget;
-      if (context_xim->client_widget)
-	{
-	  g_signal_connect (context_xim->client_widget, "hierarchy-changed",
-			    G_CALLBACK (on_client_widget_hierarchy_changed),
-			    context_xim);
-	}
-
-      update_in_toplevel (context_xim);
-    }
-#endif
+  }
 }
 
 #if NEW_GTK_IM
@@ -337,7 +339,7 @@ void add_cursor_timeout(GtkIMContextGCIN *context_xim)
 ///
 static void
 gtk_im_context_gcin_set_client_window (GtkIMContext          *context,
-				      GdkWindow             *client_window)
+                                      GdkWindow             *client_window)
 {
 #if DBG
   printf("gtk_im_context_gcin_set_client_window\n");
@@ -368,7 +370,7 @@ gtk_im_context_gcin_new (void)
 ///
 static gboolean
 gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
-				    GdkEventKey  *event)
+                                    GdkEventKey  *event)
 {
   GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
 
@@ -376,7 +378,7 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
   unsigned char *buffer = static_buffer;
 //  char *buffer = static_buffer;
   gint buffer_size = sizeof(static_buffer) - 1;
-  gint num_bytes = 0;
+  gsize num_bytes = 0;
   KeySym keysym = 0;
   Status status;
   gboolean result = FALSE;
@@ -385,7 +387,7 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
   XKeyPressedEvent xevent;
 
   xevent.type = (event->type == GDK_KEY_PRESS) ? KeyPress : KeyRelease;
-  xevent.serial = 0;		/* hope it doesn't matter */
+  xevent.serial = 0;            /* hope it doesn't matter */
   xevent.send_event = event->send_event;
   xevent.display = GDK_DRAWABLE_XDISPLAY (event->window);
   xevent.window = GDK_DRAWABLE_XID (event->window);
@@ -420,25 +422,66 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
   }
 #endif
 
+  gboolean preedit_changed = FALSE;
+  gboolean context_has_str = context_xim->pe_str && context_xim->pe_str[0];
+
   if (xevent.type == KeyPress) {
     result = gcin_im_client_forward_key_press(context_xim->gcin_ch,
       keysym, xevent.state, &rstr);
+    preedit_changed = result;
 
-#if NEW_GTK_IM
-    // this one is for mozilla, I know it is very dirty
-    if (context_xim->is_mozilla) {
-      if (context_xim->dirty_fix_off) {
-        if (result && !rstr)
-          g_signal_emit_by_name(context, "preedit_changed");
-      } else {
-        if (rstr || !result) {
-          add_cursor_timeout(context_xim);
-        }
-      }
-    }
+#if DBG
+    printf("result %d\n", result);
 #endif
 
-//    printf("jj %x %x %d %x\n", rstr, result, num_bytes, (unsigned int)buffer[0]);
+    char *tstr;
+    GCIN_PREEDIT_ATTR att[GCIN_PREEDIT_ATTR_MAX_N];
+    int cursor_pos;
+    int attN = gcin_im_client_get_preedit(context_xim->gcin_ch, &tstr, att, &cursor_pos);
+    gboolean has_str = tstr && tstr[0];
+
+    if (!context_has_str && has_str) {
+#if DBG
+      printf("emit preedit-start\n");
+#endif
+      g_signal_emit_by_name (context, "preedit-start");
+    }
+
+    if (context_has_str != has_str || (tstr && context_xim->pe_str && strcmp(tstr, context_xim->pe_str))) {
+      if (context_xim->pe_str)
+        free(context_xim->pe_str);
+      context_xim->pe_str = tstr;
+//      preedit_changed = TRUE;
+    }
+
+
+    int attsz = sizeof(GCIN_PREEDIT_ATTR)*attN;
+    if (context_xim->pe_attN != attN ||
+      context_xim->pe_att && memcmp(context_xim->pe_att, att, attsz)) {
+//      printf("att changed pe_att:%x:%d %d\n", context_xim->pe_att, context_xim->pe_attN, attN);
+      context_xim->pe_attN = attN;
+      if (context_xim->pe_att)
+        free(context_xim->pe_att);
+
+      context_xim->pe_att = NULL;
+      if (attN)
+        context_xim->pe_att = malloc(attsz);
+      memcpy(context_xim->pe_att, att, attsz);
+//      printf("context_xim->pe_att %x\n", context_xim->pe_att);
+//      preedit_changed = TRUE;
+    }
+
+    if (context_xim->pe_cursor != cursor_pos) {
+#if DBG
+      printf("cursor changed %d %d\n", context_xim->pe_cursor, cursor_pos);
+#endif
+      context_xim->pe_cursor = cursor_pos;
+//      preedit_changed = TRUE;
+    }
+
+#if DBG
+    printf("seq:%d rstr:%s result:%x num_bytes:%d %x\n", context_xim->gcin_ch->seq, rstr, result, num_bytes, (unsigned int)buffer[0]);
+#endif
     if (!rstr && !result && num_bytes && buffer[0]>=0x20 && buffer[0]!=0x7f
         && !(xevent.state & (Mod1Mask|ControlMask))) {
       rstr = (char *)malloc(num_bytes + 1);
@@ -446,13 +489,44 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
       rstr[num_bytes] = 0;
       result = TRUE;
     }
+
+#if NEW_GTK_IM
+    // this one is for mozilla, I know it is very dirty
+    if (context_xim->is_mozilla && !context_xim->dirty_fix_off) {
+        if (rstr || !result)
+          add_cursor_timeout(context_xim);
+    } else {
+//      printf("predit_changed\n");
+    }
+#endif
   }
   else {
     result = gcin_im_client_forward_key_release(context_xim->gcin_ch,
       keysym, xevent.state, &rstr);
   }
-#if 0
-  printf("event->type:%d iiiii %d  %x %d rstr:%x\n", event->type, result, keysym,
+
+
+#if 1
+  if (preedit_changed) {
+#if DBG
+    printf("preedit-change\n");
+#endif
+    g_signal_emit_by_name(context_xim, "preedit_changed");
+  }
+#endif
+
+
+  if (rstr && context_has_str) {
+    clear_preedit(context_xim);
+#if DBG
+    printf("preedit-end\n");
+#endif
+    g_signal_emit_by_name (context, "preedit-end");
+  }
+
+
+#if DBG
+  printf("seq:%d event->type:%d iiiii %d  %x %d rstr:%x\n",context_xim->gcin_ch->seq, event->type, result, keysym,
     num_bytes, rstr);
 #endif
   if (rstr) {
@@ -463,6 +537,7 @@ gtk_im_context_gcin_filter_keypress (GtkIMContext *context,
     free(rstr);
   }
 
+
   return result;
 }
 
@@ -475,7 +550,7 @@ gtk_im_context_gcin_focus_in (GtkIMContext *context)
   printf("gtk_im_context_gcin_focus_in\n");
 #endif
   if (context_xim->gcin_ch) {
-#if NEW_GTK_IM
+#if NEW_GTK_IM && 0
     if (context_xim->is_mozilla && !context_xim->dirty_fix_off)
       add_cursor_timeout(context_xim);
 #endif
@@ -495,7 +570,16 @@ gtk_im_context_gcin_focus_out (GtkIMContext *context)
 #endif
 
   if (context_xim->gcin_ch) {
-    gcin_im_client_focus_out(context_xim->gcin_ch);
+    char *rstr;
+    gcin_im_client_focus_out2(context_xim->gcin_ch , &rstr);
+
+    if (rstr) {
+      g_signal_emit_by_name (context, "commit", rstr);
+      clear_preedit(context_xim);
+      g_signal_emit_by_name(context, "preedit_changed");
+      free(rstr);
+    }
+
   }
 
   return;
@@ -504,7 +588,7 @@ gtk_im_context_gcin_focus_out (GtkIMContext *context)
 ///
 static void
 gtk_im_context_gcin_set_cursor_location (GtkIMContext *context,
-					GdkRectangle *area)
+                                        GdkRectangle *area)
 {
 #if DBG
   printf("gtk_im_context_gcin_set_cursor_location %d\n", area->x);
@@ -523,7 +607,17 @@ static void
 gtk_im_context_gcin_set_use_preedit (GtkIMContext *context,
                                     gboolean      use_preedit)
 {
-//  printf("gtk_im_context_gcin_set_use_preedit %d\n", use_preedit);
+  GtkIMContextGCIN *context_gcin = GTK_IM_CONTEXT_GCIN (context);
+//  printf("gtk_im_context_gcin_set_use_preedit %x %d\n", context_gcin->gcin_ch, use_preedit);
+  if (!context_gcin->gcin_ch)
+    return;
+  int ret;
+  if (use_preedit)
+    gcin_im_client_set_flags(context_gcin->gcin_ch, FLAG_GCIN_client_handle_use_preedit, &ret);
+  else
+    gcin_im_client_clear_flags(context_gcin->gcin_ch, FLAG_GCIN_client_handle_use_preedit, &ret);
+
+  context_gcin->preedit = use_preedit;
 }
 
 
@@ -535,11 +629,55 @@ gtk_im_context_gcin_reset (GtkIMContext *context)
 #if DBG
   printf("gtk_im_context_gcin_reset %x\n", context_gcin);
 #endif
+
+
+#if 1
+  if (context_gcin->gcin_ch) {
+    gcin_im_client_reset(context_gcin->gcin_ch);
+    if (context_gcin->pe_str && context_gcin->pe_str[0]) {
+#if DBG
+      printf("clear %x\n", context_gcin);
+#endif
+      g_signal_emit_by_name(context, "preedit_changed");
+      clear_preedit(context_gcin);
+    }
+  }
+#endif
 }
 
 /* Mask of feedback bits that we render
  */
 
+static void
+add_preedit_attr (PangoAttrList *attrs,
+		   const gchar *str,  GCIN_PREEDIT_ATTR *att)
+{
+//  printf("att %d %d\n", att->ofs0, att->ofs1);
+  PangoAttribute *attr;
+  gint start_index = g_utf8_offset_to_pointer (str, att->ofs0) - str;
+  gint end_index = g_utf8_offset_to_pointer (str, att->ofs1) - str;
+
+  if (att->flag & GCIN_PREEDIT_ATTR_FLAG_UNDERLINE)
+    {
+      attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
+      attr->start_index = start_index;
+      attr->end_index = end_index;
+      pango_attr_list_change (attrs, attr);
+    }
+
+  if (att->flag & GCIN_PREEDIT_ATTR_FLAG_REVERSE)
+    {
+      attr = pango_attr_foreground_new (0xffff, 0xffff, 0xffff);
+      attr->start_index = start_index;
+      attr->end_index = end_index;
+      pango_attr_list_change (attrs, attr);
+
+      attr = pango_attr_background_new (0, 0, 0);
+      attr->start_index = start_index;
+      attr->end_index = end_index;
+      pango_attr_list_change (attrs, attr);
+    }
+}
 
 static void
 gtk_im_context_gcin_get_preedit_string (GtkIMContext   *context,
@@ -547,19 +685,52 @@ gtk_im_context_gcin_get_preedit_string (GtkIMContext   *context,
 				       PangoAttrList **attrs,
                                        gint           *cursor_pos)
 {
+
 #if DBG
   printf("gtk_im_context_gcin_get_preedit_string %x %x %x\n", str, attrs, cursor_pos);
 #endif
-  GtkIMContextGCIN *context_xim = GTK_IM_CONTEXT_GCIN (context);
-
-  if (str)
-      *str = g_strdup ("");
 
   if (cursor_pos)
       *cursor_pos = 0;
 
   if (attrs)
       *attrs = pango_attr_list_new ();
+
+  if (!str)
+    return;
+
+#if 1
+  GtkIMContextGCIN *context_gcin = GTK_IM_CONTEXT_GCIN (context);
+  if (!context_gcin->gcin_ch) {
+empty:
+    *str=g_strdup("");
+    return;
+  }
+
+  if (cursor_pos) {
+    *cursor_pos = context_gcin->pe_cursor;
+  }
+
+  if (context_gcin->pe_str) {
+    *str=g_strdup(context_gcin->pe_str);
+  } else {
+    goto empty;
+  }
+
+#if DBG
+  printf("gtk_im_context_gcin_get_preedit_string %x attN:%d '%s'\n", context_gcin->pe_att,
+    context_gcin->pe_attN, *str);
+#endif
+  int i;
+  if (attrs)
+  for(i=0; i < context_gcin->pe_attN; i++) {
+    add_preedit_attr(*attrs, *str, &(context_gcin->pe_att[i]));
+  }
+
+#else
+  if (str)
+    *str = g_strdup("");
+#endif
 }
 
 
