@@ -14,7 +14,7 @@ typedef struct {
 extern PHOKBM phkbm;
 
 typedef struct {
-  u_char ch[CH_SZ];
+  char ch[CH_SZ];
   short count;
 } PHO_ITEM;
 
@@ -55,7 +55,8 @@ void prph(phokey_t kk);
 phokey_t pho2key(char typ_pho[]);
 gboolean save_phrase_to_db(phokey_t *phkeys, char *utf8str, int len, usecount_t usecount);
 int lookup(u_char *s);
-int find_match(char *str, int len, int *eq_N, usecount_t *usecount);
+int find_match(char *str, int *eq_N, usecount_t *usecount);
+char *phokey_to_str2(phokey_t kk, int last_number);
 
 
 #define MAX_PH_BF (90)
@@ -67,4 +68,6 @@ int find_match(char *str, int len, int *eq_N, usecount_t *usecount);
 
 #define PHO_CHAR_LEN 3
 
+#define L_BRACKET_NO 22
+#define R_BRACKET_NO 23
 #define BACK_QUOTE_NO 24
